@@ -6,21 +6,20 @@ from ..exceptions import UnexpectedToken
 from ..lexer import Token
 from ..utils import Enumerator, Serialize
 
-from .lalr_analysis import LALR_Analyzer, Shift, IntParseTable
+from .lalr_analysis import LALR_Analyzer, Shift, Reduce, IntParseTable
 
 
 ###{standalone
 class LALR_Parser(object):
     def __init__(self, parser_conf, debug=False):
-        assert all(r.options is None or r.options.priority is None
-                   for r in parser_conf.rules), "LALR doesn't yet support prioritization"
+        assert all(r.options.priority is None for r in parser_conf.rules), "LALR doesn't yet support prioritization"
         analysis = LALR_Analyzer(parser_conf, debug=debug)
-        analysis.compute_lookahead()
+        analysis.compute_lalr()
         callbacks = parser_conf.callbacks
 
         self._parse_table = analysis.parse_table
         self.parser_conf = parser_conf
-        self.parser = _Parser(analysis.parse_table, callbacks)
+        self.parser = _Parser(analysis.parse_table, callbacks, debug)
 
     @classmethod
     def deserialize(cls, data, memo, callbacks):
@@ -37,21 +36,25 @@ class LALR_Parser(object):
 
 
 class _Parser:
-    def __init__(self, parse_table, callbacks):
+    def __init__(self, parse_table, callbacks, debug=False):
         self.states = parse_table.states
-        self.start_state = parse_table.start_state
-        self.end_state = parse_table.end_state
+        self.start_states = parse_table.start_states
+        self.end_states = parse_table.end_states
         self.callbacks = callbacks
+        self.debug = debug
 
-    def parse(self, seq, set_state=None):
+    def parse(self, seq, start, set_state=None):
         token = None
         stream = iter(seq)
         states = self.states
 
-        state_stack = [self.start_state]
+        start_state = self.start_states[start]
+        end_state = self.end_states[start]
+
+        state_stack = [start_state]
         value_stack = []
 
-        if set_state: set_state(self.start_state)
+        if set_state: set_state(start_state)
 
         def get_action(token):
             state = state_stack[-1]
@@ -78,27 +81,36 @@ class _Parser:
             value_stack.append(value)
 
         # Main LALR-parser loop
-        for token in stream:
-            while True:
-                action, arg = get_action(token)
-                assert arg != self.end_state
+        try:
+            for token in stream:
+                while True:
+                    action, arg = get_action(token)
+                    assert arg != end_state
 
-                if action is Shift:
-                    state_stack.append(arg)
-                    value_stack.append(token)
-                    if set_state: set_state(arg)
-                    break # next token
-                else:
-                    reduce(arg)
+                    if action is Shift:
+                        state_stack.append(arg)
+                        value_stack.append(token)
+                        if set_state: set_state(arg)
+                        break # next token
+                    else:
+                        reduce(arg)
+        except Exception as e:
+            if self.debug:
+                print("")
+                print("STATE STACK DUMP")
+                print("----------------")
+                for i, s in enumerate(state_stack):
+                    print('%d)' % i , s)
+                print("")
+
+            raise
 
         token = Token.new_borrow_pos('$END', '', token) if token else Token('$END', '', 0, 1, 1)
         while True:
             _action, arg = get_action(token)
-            if _action is Shift:
-                assert arg == self.end_state
-                val ,= value_stack
-                return val
-            else:
-                reduce(arg)
+            assert(_action is Reduce)
+            reduce(arg)
+            if state_stack[-1] == end_state:
+                return value_stack[-1]
 
 ###}
